/*
   SmartDongle helper functions v3.2


   MicroWorks, Inc.
   2808 North Cole Road
   Boise, ID 83704

   www.smartdongle.com


   ***** BEGIN LICENSE BLOCK *****

   Version: MPL 1.1/GPL 2.0/LGPL 2.1

   The contents of this file are subject to the Mozilla Public License
   Version 1.1 (the "License"); you may not use this file except in
   compliance with the License. You may obtain a copy of the License at
   http://www.mozilla.org/MPL/

   Software distributed under the License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   for the specific language governing rights and limitations under the
   License.

   The Original Code is MicroWorks, Inc. code.

   The Initial Developer of the Original Code is MicroWorks, Inc. Portions
   created by MicroWorks, Inc. are Copyright (C) 2006
   MicroWorks, Inc. All Rights Reserved.

   Contributor(s):

   Alternatively, the contents of this file may be used under the terms of
   either the GNU General Public License Version 2 or later (the "GPL"), or
   the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   in which case the provisions of the GPL or the LGPL are applicable instead
   of those above. If you wish to allow use of your version of this file only
   under the terms of either the GPL or the LGPL, and not to allow others to
   use your version of this file under the terms of the MPL, indicate your
   decision by deleting the provisions above and replace them with the notice
   and other provisions required by the GPL or the LGPL. If you do not delete
   the provisions above, a recipient may use your version of this file under
   the terms of any one of the MPL, the GPL or the LGPL.

   ***** END LICENSE BLOCK *****
*/
#include <windows.h>
#include <setupapi.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <initguid.h>

#include "devioctl.h"
#include "usb100.h"
#include "sd_cy7c63743.h"
#include "sd_cp2.h"
#include "sd_user.h"
#include "sd_eeprom.h"
#include "sd_uskerr.h"
#include "sd_helper.h"

#define QTS_READ_RETRIES 3
#define QTS_WRITE_RETRIES 3

int sendKey(
   unsigned char opt,
   smartdongleDrvVer_t *version,
   unsigned __int64 reply,
   HANDLE *m_UsbHandle);
int getKey(unsigned __int64 *challenge, HANDLE *m_UsbHandle);

int requestHostKey(HANDLE *m_UsbHandle);
int sendChallenge(unsigned __int64 challenge, HANDLE *m_UsbHandle);

int validateUsk(
   unsigned char opt,
   unsigned __int64 P1,
   unsigned __int64 P2,
   HANDLE *m_UsbHandle);

int writePort(unsigned char port, unsigned char data, HANDLE *m_UsbHandle);
int resetSecurityState(HANDLE *m_UsbHandle);
int resetKeyState(HANDLE *m_UsbHandle);
int requestHostKey(HANDLE *m_UsbHandle);
int readKeyState(unsigned char *state, HANDLE *m_UsbHandle);

int writeEepromPage(
   unsigned char opt,
   USHORT address,
   unsigned char *page,
   HANDLE *m_UsbHandle);

int readPageFlags(unsigned char *state, HANDLE *m_UsbHandle);
int waitForPageRead(unsigned int period, BYTE *flags, HANDLE *m_UsbHandle);
int readSecurityState (unsigned char *state, HANDLE *m_UsbHandle);

int readBusyFlag (unsigned char *state, HANDLE *m_UsbHandle);
int waitForKeyState (BYTE state, unsigned int period, HANDLE *m_UsbHandle);
int readKeyState (unsigned char *state, HANDLE *m_UsbHandle);
int waitForTwc (unsigned int period, HANDLE *m_UsbHandle);
int readTwcFlag (unsigned char *state, HANDLE *m_UsbHandle);

int requestEepromPrefetch (USHORT address, HANDLE *m_UsbHandle);

int readEepromPrefetch (
   unsigned char opt,
   smartdongleDrvVer_t *version,
   unsigned char *buffer,
   int *size,
   HANDLE *m_UsbHandle);

int getSerialNumber (char *serNo, HANDLE *m_UsbHandle);
int setStrictOnly (HANDLE *m_UsbHandle);
int getOpt (unsigned char *opt, HANDLE *m_UsbHandle);
int getDriverVersion (smartdongleDrvVer_t *version, HANDLE *m_UsbHandle);

unsigned short sig;
unsigned short sig2;


/*
   Returns the first Device found, not for multiple SmartDongles.
*/
int OpenUsbDev (HANDLE *m_UsbHandle)
{
   HDEVINFO hwDeviceInfo;
   SP_DEVINFO_DATA devInfoData;
   SP_DEVICE_INTERFACE_DATA devInterfaceData;
   PSP_DEVICE_INTERFACE_DETAIL_DATA functionClassDeviceData = NULL;
   DWORD devInterface;
   ULONG predictedLength;
   ULONG requiredLength = 0;
   int status;


   /*
      Open a handle to the plug and play dev node.
      SetupDiGetClassDevs() returns a device information set that contains
      info on all installed devices of a specified class
   */
   hwDeviceInfo = SetupDiGetClassDevs (
      (LPGUID) &QTS_USK_GUID,
      NULL,
      NULL,
      DIGCF_PRESENT | DIGCF_INTERFACEDEVICE);

   if (hwDeviceInfo == INVALID_HANDLE_VALUE)
   {
      status = USK_ERR_OPEN;
      goto FAIL;
   }

   /*
      SetupDiEnumDeviceInterfaces() returns information about device
      interfaces exposed by one or more devices.  Each call returns
      information about one interface.  The routine can be called by
      one or more devices.
   */

   devInterfaceData.cbSize = sizeof (devInterfaceData);

   /* Get the FIRST interface, should be only one anyways */
   devInterface = 0;
   status = SetupDiEnumDeviceInterfaces(
      hwDeviceInfo,
      0,
      (LPGUID) &QTS_USK_GUID,
      devInterface,
      &devInterfaceData);

   if (status)
   {
      /*
         Allocate a function class device data structure to receive
          the data about this particular device.
      */
      status = SetupDiGetDeviceInterfaceDetail (
         hwDeviceInfo,
         &devInterfaceData,
         NULL,
         0,
         &requiredLength,
         NULL);

      /*
         Will always happen, looking for required length returned
         in parameter: requiredLength
      */
      if (!status)
      {
         if (GetLastError() != ERROR_INSUFFICIENT_BUFFER)
         {
            status = USK_ERR_OPEN;
            goto FAIL;
         }
      }

      predictedLength = requiredLength;

      functionClassDeviceData =
         (PSP_DEVICE_INTERFACE_DETAIL_DATA)malloc( predictedLength );

      if (functionClassDeviceData == NULL)
      {
         status = USK_ERR_OPEN;
         goto FAIL;
      }

      functionClassDeviceData->cbSize = sizeof(SP_INTERFACE_DEVICE_DETAIL_DATA);

      devInfoData.cbSize = sizeof(devInfoData);

      /* Retrieve the information from Plug and Play */
      status = SetupDiGetDeviceInterfaceDetail(
         hwDeviceInfo,
         &devInterfaceData,
         functionClassDeviceData,
         predictedLength,
         &requiredLength,
         &devInfoData);

      if (status)
      {
         /* Just open device and use default pipe 0 */
         *m_UsbHandle = CreateFile(
            functionClassDeviceData->DevicePath,
            0,
            FILE_SHARE_READ | FILE_SHARE_WRITE,
            NULL,
            OPEN_EXISTING,
            0,
            NULL
         );

         if (*m_UsbHandle == INVALID_HANDLE_VALUE)
         {
            status = USK_ERR_OPEN;
            goto FAIL;
         }
      }
      else
      {
         status = USK_ERR_OPEN;
         goto FAIL;
      }
   }
   else
   {
      status = USK_ERR_OPEN;
      goto FAIL;
   }

   status = USK_OK;

FAIL:
   if (hwDeviceInfo != INVALID_HANDLE_VALUE)
   {
      SetupDiDestroyDeviceInfoList (hwDeviceInfo);
   }

   if (functionClassDeviceData != NULL)
   {
      free (functionClassDeviceData);
   }

   return status;
}


/*
   Attempt an innocuous command.

   Return with status within dwMilliseconds.
   Time argument for device mutex access, set to zero if
   simply querying for "is this SmartDongle present?".
*/
int ready(DWORD dwMilliseconds, HANDLE *m_UsbHandle)
{
   int i;
   int retVal;
   int error;
   DWORD bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];
   PUSB_STRING_DESCRIPTOR pUsbDesc;
   unsigned char index = 0x03;  /* Index for serial number fixed at 0x03 */


   if (dwMilliseconds)
   {
      error = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
      if (error)
      {
         return error;
      }
   }

   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_STRING_DESCRIPTOR,
            &index,
            sizeof(unsigned long),
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_NODEV;
   }

   pUsbDesc = (PUSB_STRING_DESCRIPTOR)descBuf;

   /*
      If SmartDongle not plugged in, will return
      an unexpected USB descriptor string.
   */
   if ((pUsbDesc->bLength != CP2_SER_NUMB_STRING_LENGTH) ||
       (pUsbDesc->bDescriptorType != USB_STRING_DESCRIPTOR_TYPE))
   {
       return USK_ERR_NODEV;
   }

   return USK_OK;
}


/*
   Enable Dongle by stepping through the required States and using Keys.
*/
int EnableUsk(
   unsigned __int64 P1,
   unsigned __int64 P2,
   HANDLE *m_UsbHandle)
{
   unsigned __int64 challenge;
   unsigned __int64 X;
   unsigned char readErrorMask = QTSRDEEPROMERROR;
   unsigned char state;
   smartdongleDrvVer_t version;
   unsigned char opt;
   int error;


   error = getOpt(&opt, m_UsbHandle);
   if (error != USK_OK)
   {
      return error;
   }

   getDriverVersion(&version, m_UsbHandle);

   /*
       Reset key state machine here for all but v1.50 or higher SmartDongles
   */
   if (!(opt & QTS_OPTION_EXTENDED_SUPPORT))
   {
      error = resetKeyState(m_UsbHandle);
      if (error)
      {
         goto DONE;
      }

      error = waitForKeyState(QTSINITIAL, QTSSTATEWAIT, m_UsbHandle);
      if (error)
      {
         /*
            Was there a fatal EPROM read error?
            Ignore readPageFlags IOCTL error
         */
         if (readPageFlags(&state, m_UsbHandle) == USK_OK)
         {
            if (state & readErrorMask)
            {
               error = USK_ERR_EEPROM_READ;
            }
         }

         goto DONE;
      }
   }

   /*
      Request the dongle create challenge.
   */
   error = requestHostKey(m_UsbHandle);
   if (error)
   {
      goto DONE;
   }

   /* Dongle's internally generated number to challenge host ready? */
   error = waitForKeyState(QTSREPLY, QTSSTATEWAIT, m_UsbHandle);
   if (error)
   {
      /*
         Was there a fatal EPROM read error?
         Ignore readPageFlags IOCTL error
      */
      if (readPageFlags(&state, m_UsbHandle) == USK_OK)
      {
         if (state & readErrorMask)
         {
            error = USK_ERR_EEPROM_READ;
         }
      }

      goto DONE;
   }

   /*
      Get the SmartDongle's challenge.
   */
   error = getKey(&challenge, m_UsbHandle);
   if (error)
   {
      goto DONE;
   }

   X = challenge * P1 + P2;

   /*
      Reply to challenge
   */
   error = sendKey(opt, &version, X, m_UsbHandle);
   if (error)
   {
      goto DONE;
   }

   /*
      When key state sequences back to 0 (QTSINITIAL),
      the host response has been verified.
      The security state should be 1 (QTSENABLED) if the
      host reply is correct.
   */
   error = waitForKeyState(QTSINITIAL, QTSSTATEWAIT, m_UsbHandle);
   if (error)
   {
      /*
         Was there a fatal EPROM read error?
         Ignore readPageFlags IOCTL error
      */
      if (readPageFlags(&state, m_UsbHandle) == USK_OK)
      {
         if (state & readErrorMask)
         {
            error = USK_ERR_EEPROM_READ;
         }
      }

      goto DONE;
   }

   /*
      Is the host verified, dongle ENABLED?
   */
   error = readSecurityState(&state, m_UsbHandle);
   if (error)
   {
      goto DONE;
   }
   if (state != QTSENABLED)
   {
      error = USK_ERR_HOST_VALIDATE;
      goto DONE;
   }

   /*
      Force a v1.50 or greater SmartDongle to identify itself.
   */
   if (opt & QTS_OPTION_EXTENDED_SUPPORT)
   {
      error = validateUsk(opt, P1, P2, m_UsbHandle);
      if (error)
      {
         resetKeyState(m_UsbHandle);
      }
   }

DONE:
   return error;
}


/*
   Key state and Security state to 0 locks SmartDongle
*/
int DisableUsk(HANDLE *m_UsbHandle)
{
   return resetKeyState(m_UsbHandle);
}

/*
   Read from an unlocked SmartDongle, see sd_uskerr.h for returned errors.
*/
int ReadEepromUserSpace(
   unsigned __int64 P1,
   unsigned __int64 P2,
   unsigned short address,
   unsigned char *buffer,
   int *size,
   HANDLE *m_UsbHandle)
{
   unsigned char buf[QTS_IOCTL_BUFFSIZE];
   unsigned char flags = 0;
   unsigned char readErrorMask = QTSRDADDRESSRANGE|QTSRDDENY|QTSRDEEPROMERROR;
   unsigned short offset;
   unsigned short page0;

   int prefetches;
   int to_load;
   int remainder;
   int i, retries;
   int firstPrefetch;
   int singlePrefetch = 1;
   int bytesReturned;
   int prefetchSize;
   unsigned char opt;
   smartdongleDrvVer_t version;
   int error;


   error = getOpt(&opt, m_UsbHandle);
   if (error != USK_OK)
   {
      goto DONE;
   }

   getDriverVersion(&version, m_UsbHandle);


   /*
      User may request address that is inside a page.
      Get the offset from the first page.
   */
   page0 = address & ~EEPROMPAGEMASK;

   /* EPROM uses paged address: all bits below page boundaries a "don't care" */
   address &= EEPROMPAGEMASK;

   to_load = *size;
   *size = 0;

   prefetches = 1;
   prefetchSize = 0;

   for (i=0, firstPrefetch=1; i<prefetches; i++)
   {
      offset = (i*prefetchSize)+address;

      /*
         Must obtain new key for each write transaction
         if SmartDongle configured for strict access.
      */
      if ((opt & QTS_OPTION_EXTENDED_SUPPORT) &&
          (opt & QTS_OPTION_STRICT_ONLY))
      {
         error = EnableUsk(P1, P2, m_UsbHandle);
      }
      else if (firstPrefetch)
      {
         error = EnableUsk(P1, P2, m_UsbHandle);
      }

      if (error)
      {
         goto DONE;
      }

      for (retries = 0; retries<QTS_READ_RETRIES; retries++)
      {
         /* Request prefetch starting at offset */
         error = requestEepromPrefetch(offset, m_UsbHandle);
         if(error)
         {
            break;
         }

         /* Return error status if request failed */
         error = waitForPageRead(QTSPAGEWAIT, &flags, m_UsbHandle);
         if (error)
         {
            break;
         }

         /* Don't retry on an address range error */
         flags &= readErrorMask;
         if (flags)
         {
            error = USK_ERR_EEPROM_READ;

            if (flags & (QTSRDDENY|QTSRDEEPROMERROR))
            {
               continue;
            }
            else
            {
               break;
            }
         }

         /* Read buffered pages */
         error = readEepromPrefetch(
            opt,
            &version,
            buf,
            &bytesReturned,
            m_UsbHandle);

         break;
      }

      if (error)
      {
         return error;
      }

      /*
         Use the number of bytes returned calculated the number of prefetches
         needed.  Calculate using the offset into the first EPROM page.
      */
      if (firstPrefetch)
      {
         /*
            Sanity check, default to single page of prefetch.
            the 1.5 returns 1024 but has only 8 bytes of data.
         */
         if ((bytesReturned<EEPROMPAGESIZE)||(bytesReturned>QTS_MAXPREFETCH))
         {
            prefetchSize = EEPROMPAGESIZE;
         }
         else
         {
            prefetchSize = bytesReturned;
         }

         prefetches = (to_load+page0)/prefetchSize;
         remainder = (to_load+page0)%prefetchSize;
         if (remainder)
         {
            prefetches++;
         }

         /*
            A single prefetch is read with regard to page 0 offset and
            data length remainder.
         */
         if (prefetches > 1)
         {
            singlePrefetch = 0;
         }
         else
         {
            singlePrefetch = 1;
         }
      }

      /*
         Process prefetches.
      */
      if (singlePrefetch)
      {
         memcpy(buffer, buf+page0, to_load);
         *size = to_load;
      }
      else if ((to_load+page0)>=prefetchSize)
      {
         if (firstPrefetch)
         {
            memcpy(buffer, buf+page0, prefetchSize-page0);
            *size += prefetchSize-page0;
            to_load -= prefetchSize-page0;
         }
         else
         {
            memcpy(buffer+(*size), buf, prefetchSize);
            *size += prefetchSize;
            to_load -= prefetchSize;
         }
      }
      else
      {
         /*
            Load remainder.
            A "first prefetch" here would be a "single prefetch"
         */
         memcpy(buffer+(*size), buf, to_load);
         *size += to_load;
         to_load = 0;
      }

     firstPrefetch = 0;
   }

DONE:
   return error;
}


/*
   Write to an unlocked SmartDongle, see sd_uskerr.h for returned errors.
*/
int WriteEepromUserSpace(
   unsigned __int64 P1,
   unsigned __int64 P2,
   unsigned short address,
   unsigned char *buffer,
   int *size,
   HANDLE *m_UsbHandle)
{
   unsigned char buf[EEPROMPAGESIZE];
   unsigned char flags;
   unsigned char writeErrorMask = QTSWRADDRESSRANGE|QTSWRDENY|QTSWREEPROMERROR;
   unsigned short offset;
   unsigned short page0;
   int to_send, sent, received;
   int i, retries, pages;
   int firstPage;
   int singlePage;
   unsigned char opt;
   int error;


   error = getOpt(&opt, m_UsbHandle);
   if (error != USK_OK)
   {
      goto DONE;
   }


   /*
      User may request address that is inside a page. Get the offset from
      the first page.
   */
   page0 = address & ~EEPROMPAGEMASK;

   /* EPROM uses paged address: all bits below page boundaries a "don't care" */
   address &= EEPROMPAGEMASK;

   to_send = *size;
   *size = 0;

   /*
      Calculate number of write pages with respect to any offset into the
      first EPROM page.
   */
   pages = (to_send+page0)/EEPROMPAGESIZE;
   if ((to_send+page0)%EEPROMPAGESIZE)
   {
      pages++;
   }

   /*
      If a single page is written, page 0 offset and the data length remainder
      is regarded.
   */
   if (pages>1)
   {
      singlePage = 0;
   }
   else
   {
      singlePage = 1;
   }

   for (i=0, firstPage = 1; i<pages; i++)
   {
      offset = (i*EEPROMPAGESIZE)+address;

      /*
         First and last page - single page.
      */
      if (singlePage)
      {
         /*
            Retrieve current EPROM page if entire page not overwritten.
         */
         if (to_send<EEPROMPAGESIZE)
         {
            received = EEPROMPAGESIZE;
            error = ReadEepromUserSpace(
               P1, P2, offset, buf, &received, m_UsbHandle);

            /* Skip size check if IOCTL error */
            if (!error)
            {
               if (received != EEPROMPAGESIZE)
               {
                  error = USK_ERR_EEPROM_READ;
               }
            }
         }

         memcpy(buf+page0, buffer, to_send);
         sent = to_send;
      }

      /*
         Process all pages but the last page here.
      */
      else if (!(i==(pages-1)))
      {
         /*
            Address an offset into first page, then
            fetch and overwrite tail end of first page.
         */
         if ((firstPage)&&(page0))
         {
            received = EEPROMPAGESIZE;
            error = ReadEepromUserSpace(
               P1, P2, offset, buf, &received, m_UsbHandle);

            /* Skip size check if IOCTL error */
            if (!error)
            {
               if (received != EEPROMPAGESIZE)
               {
                  error = USK_ERR_EEPROM_READ;
               }
            }

            memcpy(buf+page0, buffer, (EEPROMPAGESIZE-page0));
            sent = EEPROMPAGESIZE-page0;
         }
         /*
            No offset, simply overwrite entire buffer.
         */
         else
         {
            memcpy(buf, buffer+(*size), EEPROMPAGESIZE);
            sent = EEPROMPAGESIZE;
         }
      }

      /*
         Last page.
      */
      else
      {
         /*
            Retrieve current EPROM page if entire page not to be overwritten.
         */
         if (to_send<EEPROMPAGESIZE)
         {
            received = EEPROMPAGESIZE;
            error = ReadEepromUserSpace(
               P1, P2, offset, buf, &received, m_UsbHandle);

            /* Skip size check if IOCTL error */
            if (!error)
            {
               if (received != EEPROMPAGESIZE)
               {
                  error = USK_ERR_EEPROM_READ;
               }
            }
         }

         memcpy(buf, buffer+(*size), to_send);
         sent = to_send;
      }

      if (error)
      {
         goto DONE;
      }

      /*
         Must obtain new key for each write transaction
         if SmartDongle configured for strict access.
      */
      if ((opt & QTS_OPTION_EXTENDED_SUPPORT) &&
          (opt & QTS_OPTION_STRICT_ONLY))
      {
         error = EnableUsk(P1, P2, m_UsbHandle);
      }
      else if (firstPage)
      {
         error = EnableUsk(P1, P2, m_UsbHandle);
      }

      if (error)
      {
         goto DONE;
      }

      for (retries = 0; retries<QTS_WRITE_RETRIES; retries++)
      {
         /*
            Write the page
         */
         error = writeEepromPage(opt, offset, buf, m_UsbHandle);
         if (error)
         {
            break;
         }

         error = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
         if (error)
         {
            break;
         }

         /* Any errors writing page to EEPROM? */
         error = readPageFlags(&flags, m_UsbHandle);
         if (error)
         {
            break;
         }

         /* Don't retry if an address range error */
         flags &= writeErrorMask;
         if (flags)
         {
            error = USK_ERR_EEPROM_WRITE;

            if (flags & (QTSWRDENY|QTSWREEPROMERROR))
            {
               /* Ignore error */
               waitForTwc(QTSTWCWAIT, m_UsbHandle);
               continue;
            }
            else
            {
               break;
            }
         }

         /* Wait here for Twc (Time Write cycle) to reset */
         error = waitForTwc(QTSTWCWAIT, m_UsbHandle);

         break;
      }

      if (error)
      {
         goto DONE;
      }

      to_send -= sent;
      *size += sent;

      firstPage = 0;
   }

DONE:
   return error;
}


int Green(HANDLE *m_UsbHandle)
{
   return writePort(QTSLEDPORT, QTSGREEN, m_UsbHandle);
}


int Red(HANDLE *m_UsbHandle)
{
   return writePort(QTSLEDPORT, QTSRED, m_UsbHandle);
}


int Off(HANDLE *m_UsbHandle)
{
   return writePort(QTSLEDPORT, QTSLEDOFF, m_UsbHandle);
}


/*
   Send a Reply calculated from SmartDongle's Challenge.
*/
int sendKey(
   unsigned char opt,
   smartdongleDrvVer_t *version,
   unsigned __int64 reply,
   HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   ULONG retVal;
   int i;
   int error;
   unsigned __int64 Y, X = reply;


   /*
      If signed read/write is supported
   */
   if (opt & QTS_OPTION_EXTENDED_SUPPORT)
   {
#ifdef _GNUC_
      sig = (unsigned short)((X & 0xFFFF000000000000ULL)>>48);
      sig2 = (unsigned short)((X & 0x0000FFFF00000000ULL)>>32);
#else
      sig = (unsigned short)((X & 0xFFFF000000000000)>>48);
      sig2 = (unsigned short)((X & 0x0000FFFF00000000)>>32);
#endif

      /*
         Use 32 bit signature if using >=2.3.x drivers
      */
      if (((version->major == 2) &&
          (version->minor >= 3)) ||
          (version->major >= 3))
      {
         X &= 0xFFFFFFFF;
      }
      else
      {
         /*
            Strict access requires SmartDongle driver version 2.3.0 or greater.
         */
         if (opt & QTS_OPTION_STRICT_ONLY)
         {
            error = USK_ERR_REQUIRES_DRIVER_2_3_0;
            goto DONE;
         }

         srand((unsigned)time(NULL));
         Y = (unsigned __int64)rand();
         Y <<= 52;
         X ^= Y;
      }
   }

   /* Wait for task_loop processing to finish */
   error = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
   if (error)
   {
      goto DONE;
   }

   /*
      Driver does not queue requests, retry on fail
   */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_SEND_KEY,
            &X,
            sizeof(unsigned __int64),
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      error = USK_ERR_IOCTL_SEND_KEY;
   }

DONE:
   return error;
}


/*
   Get the SmartDongle's Challenge.
*/
int getKey(unsigned __int64 *challenge, HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];
   ULONG retVal;
   int i;
   int error;


   /* Wait for task_loop processing to finish */
   error = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
   if (error)
   {
      return error;
   }

   /*
      Driver does not queue requests, retry on fail
   */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_KEY,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_GET_KEY;
   }

   memcpy(challenge, descBuf, sizeof(unsigned __int64));

   return  USK_OK;
}


/*
   SmartDongle v1.50
   Request the SmartDongle to create a Challenge.
*/
int requestHostKey(HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   ULONG retVal;
   int i;
   int error;


   /* Wait for task_loop processing to finish */
   error = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
   if (error)
   {
      return error;
   }

   /* Driver does not queue requests, retry on fail */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_SEND_HOST_VERIFY_REQUEST,
            NULL,
            0,
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_SEND_HOST_VERIFY_REQUEST;
   }

   return USK_OK;
}


/*
   Send a challenge to SmartDongle.
*/
int sendChallenge(
   unsigned __int64 challenge,
   HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   ULONG retVal;
   int i;
   int error;


   /* Wait for task_loop processing to finish */
   error = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
   if (error)
   {
      return error;
   }

   /*
      Driver does not queue requests, retry on fail
   */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_SEND_KEY,
            &challenge,
            sizeof(unsigned __int64),
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_SEND_KEY;
   }

   return USK_OK;
}


/*
   For SmartDongle v1.50, force a SmartDongle to reply to a host challenge.
   Useful if a user does not write any secrets to SmartDongle's user memory
   and simply checks for presence of a SmartDongle.  An attacker could spoof
   a protected application program by using a modified driver or usk.dll

   SmartDongle must be enabled to support a host challenge.
*/
int validateUsk(
   unsigned char opt,
   unsigned __int64 P1,
   unsigned __int64 P2,
   HANDLE *m_UsbHandle)
{
   unsigned __int64 challenge, reply, expected_reply;
   unsigned char readErrorMask = QTSRDEEPROMERROR;
   unsigned char state;
   int error;


   /*
      For v1.50 or greater SmartDongles only
   */
   if (!(opt & QTS_OPTION_EXTENDED_SUPPORT))
   {
      return USK_ERR_REQUIRES_DONGLE_1_50;
   }

   /* Make a 64 bit challenge */
   srand((unsigned)time( NULL ));
   challenge = (unsigned __int64)rand();
   challenge |= (unsigned __int64)rand() << 16;
   challenge |= (unsigned __int64)rand() << 32;
   challenge |= (unsigned __int64)rand() << 48;

   /* High bit must be set */
#ifdef _GNUC_
   challenge |= 0x8000000000000000ULL;
#else
   challenge |= 0x8000000000000000;
#endif

   /* Calculate expected reply from SmartDongle */
   expected_reply = challenge * P1 + P2;

   /*
      Send challenge.
      Assuming key state was set to 0 at termination of
      SmartDongle enable sequence.
   */
   error = sendChallenge(challenge, m_UsbHandle);
   if (error)
   {
      return error;
   }

   /* Wait for SmartDongle's reply to host challenge */
   error = waitForKeyState(QTSREPLY, QTSSTATEWAIT, m_UsbHandle);
   if (error)
   {
      /*
         Was there a fatal EPROM read error?
         Ignore readPageFlags IOCTL error
      */
      if (readPageFlags(&state, m_UsbHandle) == USK_OK)
      {
         if (state & readErrorMask)
         {
            error = USK_ERR_EEPROM_READ;
         }
      }

      return error;
   }

   /*
      Get the SmartDongle's reply to challenge.
   */
   error = getKey(&reply, m_UsbHandle);
   if (error)
   {
      return error;
   }

   if(expected_reply != reply)
   {
      return USK_ERR_USK_VALIDATE;
   }

   return USK_OK;
}


/*
   Write to LED Port
   As now: used to control indicator LED.
*/
int writePort(unsigned char port, unsigned char color, HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   ULONG retVal;
   unsigned char out[2];
   int i;


   out[0] = port;
   out[1] = color;

   /* Driver does not queue requests, retry on fail */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_WRITE_PORT,
            out,
            2*sizeof(unsigned char),
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_QTSDONGLE_SEND_PORT;
   }

   return USK_OK;
}


/*
   Get the SmartDongle's security state: 0)  Disabled
                                         1)  Enabled
*/
int readSecurityState(unsigned char *state, HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];
   ULONG retVal;
   int i;


   /*
      Driver does not queue requests, retry on fail
   */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_SECURITY_STATE,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_GET_SECURITY_STATE;
   }

    /* State should be in first byte */
    *state = descBuf[0];

    return  USK_OK;
}


/*
   Reset the SmartDongle's security state to 0  (Disabled)
*/
int resetSecurityState(HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   ULONG retVal;
   int i;


   /* Driver does not queue requests, retry on fail */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_RESET_SECURITY_STATE,
            NULL,
            0,
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_RESET_SECURITY_STATE;
   }

    return  USK_OK;
}


/*
   Get the SmartDongle's key state: 0) The initial state.
                                    1) Challenge is ready
*/
int ReadKeyState(unsigned char *state, HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];
   ULONG retVal;
   int i;


   /* Driver does not queue requests, retry on fail */
   for(i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_KEY_STATE,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
         break;

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
      return USK_ERR_IOCTL_GET_KEY_STATE;

   /* State should be in first byte */
   *state = descBuf[0];

   return USK_OK;
}


/*
   Reset the SmartDongle's key state to 0 (QTSINITIAL)
*/
int resetKeyState(HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   ULONG retVal;
   int error;
   int i;


   /*
      Wait for task_loop processing to finish
   */
   error = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
   if (error)
   {
      return error;
   }

   /*
      Driver does not queue requests, retry on fail
   */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_RESET_KEY_STATE,
            NULL,
            0,
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_RESET_KEY_STATE;
   }

   return USK_OK;
}


/*
   Send the address of a page in EEPROM that will be read
   on the next ReadEepromPrefetch.
*/
int requestEepromPrefetch(
   unsigned short address,
   HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   ULONG retVal;
   int i;


   /* Driver does not queue requests, retry on fail */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_REQUEST_PAGE,
            &address,
            sizeof(unsigned short),
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_REQUEST_PAGE;
   }

   return USK_OK;
}


/*
   Send a page of data out to the SmartDongle's write buffer, use
   readPageFlags to poll the status of the EEPROM write buffer.
*/
int writeEepromPage(
   unsigned char opt,
   unsigned short address,
   unsigned char *page,
   HANDLE *m_UsbHandle)
{
   unsigned long bytesReturned;
   unsigned long retVal;
   unsigned int length;
   unsigned char buf[4*sizeof(unsigned short) + EEPROMPAGESIZE];
   int i;


   /*
      32 bit Signature + EEPROM address + 16 bytes of page data
   */
   if (opt & QTS_OPTION_EXTENDED_SUPPORT)
   {
      memcpy(buf, &sig, sizeof(unsigned short));

      memcpy(buf+sizeof(unsigned short), &sig2, sizeof(unsigned short));

      memcpy(buf+(2*sizeof(unsigned short)), &address, sizeof(unsigned short));

      memcpy(buf+(3*sizeof(unsigned short)), page, EEPROMPAGESIZE);

      length = 4*sizeof(unsigned short) + EEPROMPAGESIZE;
   }

   /*
      EEPROM address + 16 bytes of page data
   */
   else
   {
      memcpy(buf, &address, sizeof(unsigned short));
      memcpy(buf+sizeof(unsigned short), page, EEPROMPAGESIZE);

      length = sizeof(unsigned short) + EEPROMPAGESIZE;
   }

   /*
      Driver does not queue requests, retry on fail
   */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_WRITE_EEPROM,
            buf,
            length,
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_WRITE_EEPROM;
   }

   return USK_OK;
}


/*
   Get the data currently in the SmartDongle's EEPROM page read buffer.
   Use readPageFlags to poll the status of the EEPROM read/write buffer.
*/
int readEepromPrefetch(
   unsigned char opt,
   smartdongleDrvVer_t *version,
   unsigned char *buffer,
   int *size,
   HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   unsigned char inBuf[0xfff];
   unsigned char *outBuf;
   ULONG retVal;
   unsigned int in, out;
   int i;


   /*
      If the SmartDongle supports using a "signed" USB request block.
   */
   if (opt & QTS_OPTION_EXTENDED_SUPPORT)
   {
      /*
         If using the 2.3.0 driver, then send the 32 bit
         "signature" in the output buffer.
      */
      if (((version->major == 2) &&
         (version->minor >= 3)) ||
         (version->major >= 3))
      {
         /* 32 bit Signature */
         memcpy(inBuf, &sig2, sizeof(unsigned short));
         memcpy(inBuf+sizeof(unsigned short), &sig, sizeof(unsigned short));

         in = QTS_IOCTL_BUFFSIZE;
         out = 2*sizeof(unsigned short);
         outBuf = inBuf;
      }

      /* 12 bit signature */
      else
      {
         in = (0xfff0 & sig) >> 4;

         /*
            Kludge to transmit a key when the value is under 128.
            A key value below the size of the data to be read
            will return an error from the bus driver.
         */
         if (in < (CP2_PAGE_PREFETCH_ENABLED * EEPROMPAGESIZE))
         {
            in |= 0x100;
         }

         outBuf = NULL;
         out = 0;
      }
   }
   else
   {
      in = QTS_IOCTL_BUFFSIZE;
      outBuf = inBuf;
      out = 0;
   }

   /* Driver does not queue requests, retry on fail */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_READ_EEPROM,
            outBuf,
            out,
            inBuf,
            in,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0) {
      *size = 0;
      return USK_ERR_IOCTL_READ_EEPROM;
   }

   memcpy(buffer, inBuf, bytesReturned);

   *size = bytesReturned;

   return USK_OK;
}


/*
   Wait for EPROM access status byte to indicate that requested EPROM pages are
   ready or an error has occurred.
*/
int waitForPageRead(
   unsigned int period,
   unsigned char *flags,
   HANDLE *m_UsbHandle)
{
   int i, error;
   unsigned char ready;
   int trys = 10;


   /* Page ready or error has occurred */
   ready = QTSRDPAGEVALID|QTSRDADDRESSRANGE|QTSRDDENY|QTSRDEEPROMERROR;

   for (i=0; i<trys; i++)
   {
      error = readPageFlags(flags, m_UsbHandle);
      if (error)
      {
         return error;
      }

      *flags &= ready;

      if (*flags)
      {
         return USK_OK;
      }

      Sleep(period/trys);
   }

   return USK_ERR_EEPROM_READ_TIMEOUT;
}


/*
   Get SmartDongle's EPROM access status byte.
*/
int readPageFlags(unsigned char *state, HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];
   ULONG retVal;
   int i;


   /* Driver does not queue requests, retry on fail */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_PAGE_FLAGS,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_GET_PAGE_FLAGS;
   }

   /* State should be in first byte */
   *state = descBuf[0];

   return  USK_OK;
}


/*
   Wait for busy flag to be reset, backwards compatible with <=1.40 Firmware
*/
int waitForBusy(unsigned int period, HANDLE *m_UsbHandle)
{
   BYTE data;
   int status;
   int trys = 10;
   int i;


   if (period <= 0)
   {
      return USK_ERR_UNKNOWN;
   }

   for (i=0; i<trys; i++)
   {
      status = readBusyFlag(&data, m_UsbHandle);
      if (status)
      {
         return status;
      }

      if (data == QTSREADY)
      {
         return USK_OK;
      }

      Sleep(period/trys);
   }

   return USK_ERR_BUSY_TIMEOUT;
}


int readBusyFlag(unsigned char *state, HANDLE *m_UsbHandle)
{
   int i;
   int retVal;
   DWORD bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];


   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_BUSY_FLAG,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);

   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_GET_BUSY_FLAG;
   }

   /* State should be in first byte */
   *state = descBuf[0];

   return  USK_OK;
}


/*
   Wait for a particular challenge/request state.

   Note: Removed the host challenge as in v1.41, so states are:

   initial state:  0
   challege state: v1.41:0x02  v1.42:0x01

*/
int waitForKeyState(
   unsigned char state,
   unsigned int period,
   HANDLE *m_UsbHandle)
{
   BYTE data, key_state, required_state;
   int status;
   int trys = 10;
   int i;

   #define USK_KEY_STATE_MASK  0x03


   /* boolean */
   required_state = (state && USK_KEY_STATE_MASK) ? 1: 0;


   if (period <= 0)
   {
      return USK_ERR_UNKNOWN;
   }

   for (i=0; i<trys; i++)
   {
      status = readKeyState(&data, m_UsbHandle);
      if (status)
      {
         return status;
      }

      /* boolean */
      key_state = (data && USK_KEY_STATE_MASK) ? 1: 0;

      if (key_state == required_state)
      {
         return USK_OK;
      }

      Sleep(period/trys);
   }

   return USK_ERR_KEY_STATE_TIMEOUT;
}

int readKeyState(unsigned char *state, HANDLE *m_UsbHandle)
{
   int i;
   int retVal;
   DWORD bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];


   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_KEY_STATE,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_GET_KEY_STATE;
   }

   /* State should be in first byte */
   *state = descBuf[0];

   return  USK_OK;
}


/*
   Wait Time Write Cycle flag to reset.
*/
int waitForTwc(unsigned int period, HANDLE *m_UsbHandle)
{
   BYTE data;
   int status;
   int trys = 10;
   int i;


   if (period <= 0)
   {
      return USK_ERR_UNKNOWN;
   }

   for (i=0; i<trys; i++)
   {
      status = readTwcFlag(&data, m_UsbHandle);
      if (status)
      {
         return status;
      }

      if (data == QTSREADY)
      {
         return USK_OK;
      }

      Sleep(period/trys);
   }

   return USK_ERR_TWC_RESET_TIMEOUT;
}

int readTwcFlag(unsigned char *state, HANDLE *m_UsbHandle)
{
   int i;
   int retVal;
   DWORD bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];


   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_TWC_FLAG,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
        return USK_ERR_IOCTL_GET_TWC_FLAG;
   }

   /* State should be in first byte */
   *state = descBuf[0];

   return USK_OK;
}


/*
   Reboot SmartDongle, ignore return error.
*/
int SoftReset(HANDLE *m_UsbHandle)
{
   DWORD bytesReturned;


   DeviceIoControl(
      *m_UsbHandle,
      IOCTL_QTSDONGLE_SOFT_RESET,
      NULL,
      0,
      NULL,
      0,
      &bytesReturned,
      NULL);

   /* After SmartDongle reset, handle becomes invalid and can not be closed. */
   // CloseHandle(m_UsbHandle);

   m_UsbHandle = INVALID_HANDLE_VALUE;

    return  USK_OK;
}


/*
   Set SmartDongle eprom mode bits.

   This is the encryption for the data on the SmartDongle
   that protects the EEPROM memory from a physical attack.

   Modes:   0x00 QTSSTRONGCRYPT    AES Encryption
            0x01 QTSCRYPTDISABLED  No Encryption
            0x02 QTSFASTCRYPT      Fast, lightwieght encryption.

   All other bits are "don't care"
*/
int SetMode(
   unsigned __int64 P1,
   unsigned __int64 P2,
   unsigned char mode,
   HANDLE *m_UsbHandle)
{
   int i;
   int retVal;
   DWORD bytesReturned;
   unsigned char opt;
   unsigned char currentMode;
   int error;


   /* Mask out extraneous bits */
   mode &= QTS_OPTION_ON_CHIP_EPROM_PROTECT_DISABLE|
            QTS_OPTION_ON_CHIP_EPROM_PROTECT_FAST;

   error = getOpt(&opt, m_UsbHandle);
   if (error != USK_OK)
   {
      return error;
   }

   /*
      v1.50 SmartDongle must be reset if mode change
   */
   if (!(opt & QTS_OPTION_EXTENDED_SUPPORT2) &&
       (opt & QTS_OPTION_EXTENDED_SUPPORT))
   {
      error = GetMode (&currentMode, m_UsbHandle);
      if (error != USK_OK)
      {
         return USK_ERR_IOCTL_SET_MODE;
      }

      /*
         Change mode only if current mode
         mismatches requested mode.
      */
      if (mode != currentMode)
      {
         SoftReset(m_UsbHandle);

         Sleep(QTS_PLUG_N_PLAY_WAIT);

         for (i=0; i<QTS_READY_RETRIES; i++)
         {
            error = OpenUsbDev(m_UsbHandle);
            if (error != 0)
            {
               Sleep(QTS_READY_WAIT);
               continue;
            }

            error = ready(0, m_UsbHandle);
            if (error == 0)
            {
               break;
            }

            if (m_UsbHandle != INVALID_HANDLE_VALUE)
            {
               CloseHandle(m_UsbHandle);
               m_UsbHandle = INVALID_HANDLE_VALUE;
            }

            Sleep(QTS_READY_WAIT);
         }

         if (error)
         {
            return USK_RESET_COMEREADY;
         }
      }

      /*
          Requested mode same as current mode, return.
      */
      else
      {
         return USK_OK;
      }
   }

   error = EnableUsk(P1, P2, m_UsbHandle);
   if (error != USK_OK)
   {
      return USK_ERR_IOCTL_SET_MODE;
   }

   /* Driver does not queue requests, retry on fail */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_SET_MODE,
            &mode,
            sizeof(unsigned char),
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      error = USK_ERR_IOCTL_SET_MODE;
   }

   /* Disable SmartDongle */
   resetKeyState(m_UsbHandle);

   return error;
}


/*
   Read SmartDongle's current eprom bits.

   This is the encryption for the data on the SmartDongle memory that protects
   the memory from a physical attack.

   Modes:   0x00 QTSSTRONGCRYPT    AES Encryption
            0x01 QTSCRYPTDISABLED  No Encryption
            0x02 QTSFASTCRYPT      Fast, lightwieght encryption.
*/
int GetMode(unsigned char *mode, HANDLE *m_UsbHandle)
{
   int i;
   int retVal;
   DWORD bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];
   int status;


   /*
      Wait for task_loop processing to finish
   */
   status = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
   if (status != USK_OK)
   {
      return status;
   }

   /* Driver does not queue requests, retry on fail */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_MODE,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_GET_MODE;
   }

   /* State should be in first byte */
   *mode = descBuf[0];

   /* Extraneous bits not masked out in firmware */
   *mode &= (QTS_OPTION_ON_CHIP_EPROM_PROTECT_DISABLE|QTS_OPTION_ON_CHIP_EPROM_PROTECT_FAST);

   return USK_OK;
}


/*
   Get SmartDongle serial number.
*/
int getSerialNumber(char *serNo, HANDLE *m_UsbHandle)
{
   int i;
   int retVal;
   int status;
   DWORD bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];
   PUSB_STRING_DESCRIPTOR pUsbDesc;
   unsigned char index = 0x03;  /* Index for serial number fixed at 0x03 */
   char *tmp;


   tmp = (char*)malloc(6*EEPROMPAGESIZE);
   if (tmp == NULL)
   {
      status = USK_ERR_MEM;
      goto DONE;
   }

   /* Wait for task_loop processing to finish */
   status = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
   if (status != USK_OK)
   {
      goto DONE;
   }

   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_STRING_DESCRIPTOR,
            &index,
            sizeof(unsigned long),
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
         break;

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      status = USK_ERR_IOCTL_GET_STRING_DESCRIPTOR;
      goto DONE;
   }

   pUsbDesc = (PUSB_STRING_DESCRIPTOR)descBuf;

   /* Returns a Unicode string and 2 UCHARS */
   if (pUsbDesc->bLength <((CP2_SERIALNUM_LENGTH * 2)+2))
   {
      status = USK_ERR_DATA_FORMAT;
      goto DONE;
   }

   /* Ignore second byte of Unicode string. */
   memcpy(tmp, pUsbDesc->bString, (CP2_SERIALNUM_LENGTH*2));

   /* An Unicode to ASCII conversion sufficient for our purpose */
   for (i = 0; i< CP2_SERIALNUM_LENGTH; i++)
   {
      serNo[i] = tmp[i*2];
   }

   serNo[i] = '\0';


DONE:
   if (tmp)
   {
      free(tmp);
   }

   return status;
}


/*
   For SmartDongle >=v1.50 using >= v2.3 driver only.
   This will configure the SmartDongle for Strict Access.
   A new key and 32 bit signatures are required for each user
   read/write transaction against user memory.

   Caution: This configuration is permanent.
            SmartDongle must be sent back to the manufacturer to
            be reset to use the older drivers (<= v2.2.0).
*/
int setStrictOnly(HANDLE *m_UsbHandle)
{
   int i;
   int retVal;
   DWORD bytesReturned;
   smartdongleDrvVer_t version;
   unsigned char opt;
   int error;


   error = getOpt(&opt, m_UsbHandle);
   if (error != USK_OK)
   {
      return error;
   }

   /*
      Requires >=1.50 SmartDongles
   */
   if (!(opt & QTS_OPTION_EXTENDED_SUPPORT))
   {
      error = USK_ERR_REQUIRES_DONGLE_1_50;
      goto DONE;
   }

   /*
      Requires >=2.3.x drivers
   */
   getDriverVersion(&version, m_UsbHandle);
   if (!(((version.major == 2) &&
      (version.minor >= 3)) ||
      (version.major >= 3)))
   {
      error = USK_ERR_REQUIRES_DRIVER_2_3_0;
      goto DONE;
   }

   /*
      Keep customer from burning memory with repeated requests.
      Return if SmartDongle previously set to strict access.
   */
   if (opt & QTS_OPTION_STRICT_ONLY)
   {
      goto DONE;
   }

   /* Wait for task_loop processing to finish */
   error = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
   if (error)
   {
      goto DONE;
   }

   /*
      Driver does not queue requests, retry on fail
   */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_SET_STRICT,
            NULL,
            0,
            NULL,
            0,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      error = USK_ERR_IOCTL_QTSDONGLE_SET_STRICT;
   }

DONE:
   return error;
}


/*
   Read SmartDongle's extended functionality bits are mapped with
   eprom error bits.
*/
int getOpt(unsigned char *opt, HANDLE *m_UsbHandle)
{
   int i;
   int retVal;
   DWORD bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];
   int error;
   unsigned char fixedOptions;

   /* Firmware support bitmap */
   #define QTS_FIRMWARE_EXTENDED_SUPPORT2  0x40
   #define QTS_FIRMWARE_EXTENDED_SUPPORT  0x08
   #define QTS_FIRMWARE_STRICT_ONLY  0x04


   /* clear options */
   *opt = 0;

   /*
      Wait for task_loop processing to finish
   */
   error = waitForBusy(QTSBUSYWAIT, m_UsbHandle);
   if (error)
   {
      return error;
   }

   /*
      Driver does not queue requests, retry on fail
   */
   for (i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_EEPROM_ERROR_FLAG,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   if (retVal == 0)
   {
      return USK_ERR_IOCTL_GET_EEPROM_ERROR_FLAG;
   }

   /* State should be in first byte */
   fixedOptions = descBuf[0];


   /*
      Map firmware flags to options flags
   */

   if (fixedOptions & QTS_FIRMWARE_EXTENDED_SUPPORT2)
   {
      *opt |= QTS_OPTION_EXTENDED_SUPPORT2;
   }

   if (fixedOptions & QTS_FIRMWARE_EXTENDED_SUPPORT)
   {
      *opt |= QTS_OPTION_EXTENDED_SUPPORT;
   }

   if (fixedOptions & QTS_FIRMWARE_STRICT_ONLY)
   {
      *opt |= QTS_OPTION_STRICT_ONLY;
   }

   return USK_OK;
}


/*
   Some functionality only supported in new, unsigned driver.

   ioBuffer[0] =  QTSDRIVER_VERSION & 0x000000FF;        // m3
   ioBuffer[1] = (QTSDRIVER_VERSION & 0x0000FF00) >> 8;  // m2
   ioBuffer[2] = (QTSDRIVER_VERSION & 0x00FF0000) >> 16; // minor
   ioBuffer[3] = (QTSDRIVER_VERSION & 0xFF000000) >> 24; // major
*/
int getDriverVersion (smartdongleDrvVer_t *version, HANDLE *m_UsbHandle)
{
   ULONG bytesReturned;
   unsigned char descBuf[QTS_IOCTL_BUFFSIZE];
   ULONG retVal;
   int i;


   /*
      Driver does not queue requests, retry on fail
   */
   for(i=0;i<QTS_IOCTL_RETRYS; i++)
   {
      retVal =
         DeviceIoControl(
            *m_UsbHandle,
            IOCTL_QTSDONGLE_GET_DRIVER_REVISION,
            NULL,
            0,
            descBuf,
            QTS_IOCTL_BUFFSIZE,
            &bytesReturned,
            NULL);

      if (retVal)
      {
         break;
      }

      Sleep(QTS_IOCTL_FAIL_WAIT);
   }

   /*
      fall back to lowest common denominator
   */
   if (retVal == 0)
   {
      version->m3 = 0x00;
      version->m2 = 0x00;
      version->minor = 0x02;
      version->major = 0x02;

      return USK_ERR_IOCTL_GET_DRIVER_REVISION;
   }

   memcpy(version, descBuf, sizeof(smartdongleDrvVer_t));

   return USK_OK;
}
