/*
   sd_uskerr.h for SmartDongle API version 3.2


   www.smartdongle.com

   MicroWorks, Inc.
   2808 North Cole Road
   Boise, ID 83704
*/

#ifndef _DEFINED_84C05DBE_988B_4FD9_AC5C_BF7E09646084
#define _DEFINED_84C05DBE_988B_4FD9_AC5C_BF7E09646084

#if defined(__cplusplus) || defined(__cplusplus__)
extern "C" {
#endif

#define USK_OK                                   0

/* Can't open dongle */
#define USK_ERR_OPEN                             1

/* Dongle is already open.*/
#define USK_ERR_IS_OPEN                          2

/* Memory allocation error */
#define USK_ERR_MEM                              3

/* Deprecated */
#define USK_ERR_EEPROM_SIZE                      4

/* Reserved */
#define USK_ERR_5                                5

/* Timeout waiting for state change */
#define USK_ERR_SECURITY_STATE_TIMEOUT           6

/* Timeout waiting for state change */
#define USK_ERR_KEY_STATE_TIMEOUT                7

/* Timeout waiting for TWC to reset */
#define USK_ERR_TWC_RESET_TIMEOUT                8

/* Timeout waiting for task_loop to be flagged as inactive.*/
#define USK_ERR_BUSY_TIMEOUT                     9

/* The Dongle is Not in ENABLED Mode */
#define USK_ERR_NOT_ENABLED                     10

/* Timeout Waiting For Read Flag(s) to be set */
#define USK_ERR_EEPROM_READ_TIMEOUT             11

/* An EEPROM Read ERROR Occurred */
#define USK_ERR_EEPROM_READ                     12

/* An EEPROM Write ERROR Occurred */
#define USK_ERR_EEPROM_WRITE                    13

/* Dongle failed to validate against user keys */
#define USK_ERR_USK_VALIDATE                    14

/* Host failed to validate against dongle keys */
#define USK_ERR_HOST_VALIDATE                   15

/* An error occurred when attempting to lock SmartDongle */
#define USK_ERR_DISABLE                         16

/* Data format Error */
#define USK_ERR_DATA_FORMAT                     17

/* IOCTL's Failed: see sd_user.h */
#define USK_ERR_IOCTL_GET_PIPE_INFO             18
#define USK_ERR_IOCTL_GET_DEVICE_DESCRIPTOR     19
#define USK_ERR_IOCTL_GET_CONFIG_DESCRIPTOR     20
#define USK_ERR_IOCTL_REGISTER_NOTIFY_EVENT     21
#define USK_ERR_IOCTL_RESET_DEVICE              22
#define USK_ERR_IOCTL_RESET_PIPE                23
#define USK_ERR_IOCTL_GET_DRIVER_REVISION       24
#define USK_ERR_IOCTL_GET_STRING_DESCRIPTOR     25
#define USK_ERR_26                              26 /* Deprecated */
#define USK_ERR_IOCTL_SEND_KEY                  27
#define USK_ERR_IOCTL_GET_KEY                   28
#define USK_ERR_IOCTL_WRITE_PORT                29
#define USK_ERR_IOCTL_READ_PORT                 30
#define USK_ERR_IOCTL_READ_EEPROM               31
#define USK_ERR_IOCTL_WRITE_EEPROM              32
#define USK_ERR_IOCTL_GET_SECURITY_STATE        33
#define USK_ERR_IOCTL_RESET_SECURITY_STATE      34
#define USK_ERR_IOCTL_GET_KEY_STATE             35
#define USK_ERR_IOCTL_RESET_KEY_STATE           36
#define USK_ERR_IOCTL_SEND_HOST_VERIFY_REQUEST  37
#define USK_ERR_IOCTL_REQUEST_PAGE              38
#define USK_ERR_IOCTL_GET_PAGE_FLAGS            39
#define USK_ERR_IOCTL_GET_TWC_FLAG              40
#define USK_ERR_IOCTL_GET_BUSY_FLAG             41
#define USK_ERR_IOCTL_GET_EEPROM_ERROR_FLAG     42

/* 43, 44 Reserved */
#define USK_ERR_43                              43
#define USK_ERR_44                              44

/*
   Firmware v1.41 and above.
*/
#define USK_ERR_IOCTL_QTSDONGLE_SEND_PORT       45
#define USK_ERR_IOCTL_QTSDONGLE_GET_FLAGS       46
#define USK_ERR_IOCTL_SET_MODE                  47
#define USK_ERR_IOCTL_GET_MODE                  48

/*
   For multiple SmartDongles, no device found.
   If argument specifies serial number, SmartDongle with
   the given serial number not found, else no SmartDongles
   are on the host.
*/
#define USK_ERR_NODEV                           49

#define USK_ERR_MUTEX_DEV_CREATE                54
#define USK_ERR_MUTEX_DEV_TIMEOUT               55
#define USK_ERR_MUTEX_DEV_ABANDONED             56
#define USK_ERR_MUTEX_DEV_UNKNOWN               57

#define USK_ERR_MUTEX_API_CREATE                58
#define USK_ERR_MUTEX_API_TIMEOUT               59
#define USK_ERR_MUTEX_API_ABANDONED             60
#define USK_ERR_MUTEX_API_UNKNOWN               61

/* One or more arguments are not correct */
#define USK_ERR_BAD_ARGUMENTS                   63

/* Error creating mutex Access Control List */
#define USK_ERR_SID_RETRIEVE                    65
#define USK_ERR_ACL_INIT                        66
#define USK_ERR_ACE_ADD                         67
#define USK_ERR_SD_INIT                         68
#define USK_ERR_SD_SET                          69


/* For SmartDongle v1.50 */
/* 80 Reserved */
#define USK_ERR_80                              80
#define USK_ERR_IOCTL_QTSDONGLE_SET_STRICT      81

/*
   An attempt to use extended driver functionality with old driver.
*/
#define USK_ERR_REQUIRES_DRIVER_2_3_0           82

/*
   An attempt to use extended device functionality with old SmartDongle.
*/
#define USK_ERR_REQUIRES_DONGLE_1_50            83

/*
   A warning that data written to SmartDongle after a mode change was not on
   16 byte boundaries. (i.e. 0x00, 0x10, 0x20 etc.).

   Ignore this error if all data written on page where boundary violaton
   occurred is written with SmartDongle set to the same mode.

   ReadX() and WriteX() user options:

      QTS_OPTION_ON_CHIP_EPROM_PROTECT_DISABLE
      QTS_OPTION_ON_CHIP_EPROM_PROTECT_FAST
      QTS_OPTION_ON_CHIP_EPROM_PROTECT_STRONG
*/
#define USK_WARN_MODE_PAGE_BOUNDARY             84

/*
   A warning that encrypted data written to SmartDongle was
   not on 16 byte boundaries. (i.e. 0x00, 0x10, 0x20 etc.).

   If if the user opts for AES encrytion when writing to all of
   user space, this will not be a problem.

   ReadX() and WriteX user option:

      QTS_OPTION_CRYPT_IO
*/
#define USK_WARN_CRYPT_PAGE_BOUNDARY            85

#define USK_RESET_COMEREADY                     86


#define USK_ERR_UNKNOWN                        255

#if defined(__cplusplus) || defined(__cplusplus__)
}
#endif

#endif /* _DEFINED_84C05DBE_988B_4FD9_AC5C_BF7E09646084 */
